//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef EVENT_H
#define EVENT_H

#include <sigc++/sigc++.h>
#include <string>
#include "xmlhelper.h"
#include "Reaction.h"

/**
  * Class for events which do something during a scenario.
  *
  * This is the base class for events.
  * There are different events for different occassions. Each event connects to
  * one of the several signals which are raised when something happens. Such an
  * event can trigger actions which can be anything in the range from winning
  * the game to gaining or loosing some troops.
  *
  * The correct type of the constructor etc. can only be determined in the
  * derived classes, since they will want to be fed with different data.
  *
  * The classes up to now are:
  * EPlayerDead:    raised when the specified player dies
  * EKillAll:       raised when all players except one have died
  * ECityConq:      raised when the specified city is conquered by a player
  *   \-> EPCityConq:     raised when the specified city is conquered by the 
  *                       specified player (derived from ECityConq)
  * EArmyKilled:    raised when a specified army has been killed.
  * ERound:         raised at the beginning of a specific round
  *   \-> EPRound:  raised at the beginning of a specific player's specific turn
  * ERuinSearch:    raised when a specific ruin has been searched
  */

class Event : public SigC::Object
{
    public:
        enum TYPE {KILLALL, PLAYERDEAD, CITYCONQUERED, PCITYCONQUERED, ARMYKILLED,
                    ROUND, PROUND, RUINSEARCH};
    
        //! Creates an event with type as its, well, type
        Event(TYPE type);

        //! Does the basic loading functionality, especially loading of reactions
        Event(XML_Helper* helper);

        virtual ~Event();

        /** Load an event
          * 
          * @param helper the XML_Helper object representing the savegame
          *
          * This function examines the event type and loads the correct event
          * with this information.
          */
        static Event* loadEvent(XML_Helper* helper);

        //! Callback for loading of the reactions. See XML_Helper for details
        bool loadReaction(std::string tag, XML_Helper* helper);
        
        /** Saves the event data. The base class only saves the common data!
          * Opening tags and such has to be done by the derived classes.
          */
        virtual bool save(XML_Helper* helper);

        /** Initialises the event. This is neccessary because events are loaded
          * first, but usually hook up to signals of other classes which are
          * loaded only after the events have been loaded.
          *
          * The init function is called automatically after a savegame has been
          * loaded by GameScenario.
          */
        virtual void init() {}

        //! Returns the type of the event
        TYPE getType() {return d_type;}
        
        //! Returns false if the event is processed and can be deleted.
        bool getActive() {return d_active;}

        //! Sets the active value of the event to false
        void deactivate() {d_active = false;}

        //! Add a reaction to be triggered when this event is raised
        void addReaction(Reaction* r);

        //! Get the ith reaction or zero if this one doesn't exist
        Reaction* getReaction(int index);

        //! triggers all reactions
        void raise();

    protected:
        //Data
        std::list<Reaction*> d_reactions;
        TYPE d_type;
        bool d_active;
};

#endif //EVENT_H
