//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

/* This is the GraphicsCache class. As the name suggests it is here to cache
 * already prepared graphics.
 * 
 * With the new player-specific colors, each time a stack picture is needed, it
 * has to be constructed from a kind of basic stack picture and an overlying
 * mask for the player-specific color regions. This procedure makes the units
 * look cooler, but needs some additional blits. To reduce the overhead from
 * these blits, the graphics are cached in this class. When an adjustable
 * maximum size is reached, the graphics whose use dates back the longest time
 * are deleted.
 *
 * @note For efficiency reasons, the class will not copy the surfaces it has,
 * so DON'T MODIFY THEM unless you know what you do.
 * 
 * About the internals:
 * The pictures are stored in a list of GraphicsCacheItems, which hold the name
 * of the armyset, the army number, the player and the surface itself. When a
 * picture is requested, the class first looks in its cache. If it doesn't find
 * the pic there, it loads it and deletes older pics if the maximum memory
 * consumption has been exceeded.
 */

#ifndef GRAPHICS_CACHE_H
#define GRAPHICS_CACHE_H

#include <paragui.h>
#include <string>
#include <list>

#include "player.h"

struct ArmyCacheItem;
struct CityCacheItem;

/** Soliton class for caching army and map images
  * 
  * With the introduction of player-specific colors, the problem of caching
  * images has popped up. The player colors are implemented by taking an army
  * picture and a mask, with the mask being a 16 color image, and substituting
  * the colors in the mask and blitting the mask over the army (or e.g. city)
  * image. This takes several (AFAIR three) blits and is therefore costly.
  *
  * This class approaches this problem by keeping formatted images. You get
  * e.g. an army image by querying the cache, which either gives you the cached
  * image or creates a new one. If the size is exceeded, the class will free
  * the oldest (not used for the longest time) images until it gets below the
  * treshold. It maintains some kind of balance between cached city pictures
  * and cached army pictures, but the kind of balance may change, so I'll
  * better not describe it here. :)
  *
  * The maximum cache size can be changed on the fly by modifying the value in 
  * the Configuration class.
  */

class GraphicsCache
{
    public:
        //! Function for getting/creating the soliton instance.
        static GraphicsCache* getInstance();

        //! Explicitely deletes the soliton instance
        static void deleteInstance();

        //! Frees the whole cache.
        static void clearInstance();

        //! Get the maximum cache size.
        static Uint32 getCacheSize();

        /** Function for getting the army picture from the cache
          * 
          * This function returns either the cached image of the given type or
          * creates a new one and caches it. Use this function to access
          * army images! And: Don't touch the returned surface!! For performance
          * reasons you get the original surface which is also in the cache.
          *
          * The returned surface contains the correct player colors and some
          * icons displaying the level of the unit.
          *
          * @param armyset      the armyset to be used
          * @param army         the index of the army to be used
          * @param player       the player owning the army
          * @param level        the level of the unit
          * @return the image of the unit
          */
        SDL_Surface* getArmyPic(std::string armyset, Uint32 army, Player* p,int level);

        /** Function for getting a city picture
          * 
          * For simplicity we have extended the basic_image/mask style to
          * cities as well, since it greatly reduces the number of images.
          * Use this function solely to get city images, and don't touch the
          * images!
          *
          * @param type         the level of the city
          * @param player       the player owning the city
          * @return image of the described city
          */
        SDL_Surface* getCityPic(int type, Player* p);

    private:
        GraphicsCache();
        ~GraphicsCache();

        //! Checks if the cache has exceeded the maximum size and reduce it.
        void checkPictures();
        
        //! Creates a new army picture with the given parametres.
        ArmyCacheItem* addArmyPic(std::string armyset, Uint32 army, Player* p, int level);

        //! Creates a new city picture with the given parametres.
        CityCacheItem* addCityPic(int type, Player* p);

        //! Clears the whole cache.
        void clear();

        //! Erases the oldest (least recently requested) army cache item.
        void eraseLastArmyItem();

        //! Erases the oldest (least recently requested) city cache item.
        void eraseLastCityItem();

        //! Loads the images for the city pictures and their masks.
        void loadCityPics();
        
        //the data
        static GraphicsCache* s_instance;
        static SDL_Surface* s_levelmask;

        Uint32 d_armysize;
        Uint32 d_citysize;
        std::list<ArmyCacheItem*> d_armylist;
        std::list<CityCacheItem*> d_citylist;

        //some private surfaces
        SDL_Surface* d_citypic[3];
        SDL_Surface* d_citymask[3];
};

#endif
