//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "QuestsManager.h"
#include "QRuinSearch.h"
#include "QKillHero.h"
#include "QEnemyArmies.h"
#include "stacklist.h"
#include "QuestCompletedDialog.h"

QuestsManager* QuestsManager::s_instance = NULL;

using namespace std;

//#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
#define debug(x)
//======================================================================
QuestsManager::QuestsManager()
{
    _sharedInit();
}
//======================================================================
void QuestsManager::_sharedInit()
{
    debug("QuestsManager constructor: connecting to the signals")
    sdyingArmy.connect( SigC::slot(*this, &QuestsManager::_dyingArmy));
    sendingTurn.connect( SigC::slot(*this, &QuestsManager::_cleanup));

    // now prepare the vector of pointers to the
    // functions (class static members) checking feasibility
    // for every quest
    d_questsFeasible.push_back(&(QuestRuinSearch::isFeasible));
    d_questsFeasible.push_back(&(QuestKillHero::isFeasible));
    d_questsFeasible.push_back(&(QuestEnemyArmies::isFeasible));
}
//======================================================================
QuestsManager::~QuestsManager()
{
    QuestsHashType::iterator it;

    for (it = d_quests.begin(); it != d_quests.end(); it++) 
        delete (*it).second.quest;
    _cleanup();
}
//======================================================================
QuestsManager::QuestsManager(XML_Helper* helper)
{
    _sharedInit();
    debug("QuestsManager: registerTag!");
    helper->registerTag("quest", SigC::slot((*this), &QuestsManager::load));
}
//===========================================================================
bool QuestsManager::load(string tag, XML_Helper* helper)
{
    debug("QuestsManager: load tag = " << tag);
    if (tag == "quest")
    {
        unsigned int questType;
        int          heroId;
        helper->getData(questType, "type");
        helper->getData(heroId,    "hero");

        debug("quest load: type = " << questType << ", heroId = " << heroId);

        Quest *quest=0;

        switch (questType) {
            case 1: 
                quest = new QuestRuinSearch(*this, heroId, helper);
            break;
            case 2:
                quest = new QuestKillHero(*this, heroId, helper);
                break;
            break;
            case 3:
                quest = new QuestEnemyArmies(*this, heroId, helper);
                break;
            break;
        }
        debug("quest created: q = " << quest);
        if (quest)
        {
            QRecordType rec = { questType, quest };
            d_quests[heroId] = rec;
        }
        debug("QuestsManager: return true");
        return true;
    }
    return false;
}
//======================================================================
QuestsManager* QuestsManager::getInstance(void)
{
    if (s_instance == 0)
        s_instance = new QuestsManager();
    return s_instance;
}
//======================================================================
QuestsManager* QuestsManager::getInstance(XML_Helper* helper)
{
    if (s_instance)
        deleteInstance();

    s_instance = new QuestsManager(helper);
    return s_instance;
}
//======================================================================
bool QuestsManager::save(XML_Helper* helper)
{
	bool retval = true;
	retval &= helper->openTag("questlist");
    QuestsHashType::iterator it;
    debug("Saving quests\n");
    for (it = d_quests.begin(); it != d_quests.end(); it++) 
    {
	    retval &= helper->openTag("quest");
	    retval &= helper->saveData("type", (*it).second.questType);
	    retval &= helper->saveData("hero", (*it).first);
        // save the quest:
	    retval &= ((*it).second.quest)->save(helper);
	    retval &= helper->closeTag();
    }
    debug("Quests saved\n");
	retval &= helper->closeTag();
	return retval;
}
//=======================================================================
void QuestsManager::deleteInstance()
{
    debug("QuestsManager: deleteInstance")
    delete s_instance;
    s_instance = 0;
}
//========================================================================
void QuestsManager::_dyingArmy(Army* army, culpritsType culprits)
{
    if (army->isHero() == 0)
        return;
    int id = army->getId();
    if (d_quests.count(id))
    {
        debug("QuestsManager: the hero is dead, clearing his quests...");
        _deactivateQuest(id);
    }
}
//======================================================================
void QuestsManager::_cleanup(Player::TYPE type)
{
    debug("QuestsManager: _cleanup!");

    while (d_inactive_quests.empty() == false)
    {
        delete d_inactive_quests.front();
        d_inactive_quests.pop();
    }
}
//======================================================================
Quest* QuestsManager::createNewQuest(int heroId)
{
    Quest *quest = NULL;
    // don't let a hero have more than one quest
    if (d_quests.count(heroId))
        return NULL;

    int which = 0;
    while (which == 0)
    {
        which = 1 + rand() % 3;
        // if this quest is not feasible - try again with another
        // quest:
        if ((*(d_questsFeasible[which-1]))() == 0)
            which = 0;
    }
    // ok - this quest can be completed
    switch (which)
    {
        case 1:
            quest = new QuestRuinSearch(*this, heroId);
            break;
        case 2:
            quest = new QuestKillHero(*this, heroId);
            break;
        case 3:
            quest = new QuestEnemyArmies(*this, heroId);
            break;
    }
    if (quest)
    {
        QRecordType rec = { which, quest };
        d_quests[heroId] = rec;
    }
    return quest;
}
//========================================================================
int QuestsManager::questCompleted(int heroId)
{
    /* show the dialog window */
    debug("QuestCompleted dialog");
    QuestCompletedDialog dlg(NULL, d_quests[heroId].quest);
	dlg.Show();
	dlg.RunModal();
    dlg.Hide();
    debug("deactivate quest");
    _deactivateQuest(heroId);
    debug("quest deactivated");
    return 0;
}
//========================================================================
void QuestsManager::_deactivateQuest(int heroId)
{
    Quest *q = d_quests[heroId].quest;
    q->deactivate();
    d_inactive_quests.push(q);
    // delete it from hash of active quests
    d_quests.erase(heroId);
}
//=========================================================================
void QuestsManager::getPlayerQuests(Player *player, vector<Quest*>& dst,
                                    vector<Hero*>& heroes)
{
    // loop through the player's units
    // for every hero check any pending quests
    Stacklist* stacklist = player->getStacklist();
    for (Stacklist::const_iterator it = stacklist->begin(); 
         it != stacklist->end(); it++)
    {
        for (Stack::iterator sit = (*it)->begin(); 
             sit != (*it)->end(); sit++)
	    {
            int heroId = (*sit)->getId();
            if (((*sit)->isHero()) && (d_quests.count(heroId)))
            {
                debug("heroId = " << heroId << " - has quest: " 
                      << d_quests[heroId].quest);
                dst.push_back(d_quests[heroId].quest);
                // since it IS a hero, conversion must succeed!
                heroes.push_back(dynamic_cast<Hero*>(*sit));
            }
        }
    }
}
