//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef CITY_H
#define CITY_H

#include <string>
#include "armyset.h"
#include "Location.h"
#include "xmlhelper.h"

class Player;
class Stack;
class Hero;


typedef struct
{
    std::string armyset;
    int type;
} CityProduction;

/** A city on the game map
  * 
  * The City class includes the whole functionality of a city on the game map. A
  * city can produce armies, place produced armies on its fields and buy
  * production as well as be upgraded.
  */


class City : public Location
{
    public:
        /** Preferred constructor
          * 
          * @param pos          the location of the city
          * @param name         the name of the city
          * @param gold         the amount of gold the city produces each turn
          */
        City(PG_Point pos, std::string name = "Noname", int gold = 30);

        //! The loading constructor. See XML_Helper for details.
        City(XML_Helper* helper);
        ~City();

        
        //! Set the player owning the city
        void setPlayer(Player* player){d_player = player;}

        //! Set the gold the city produces each turn
        void setGold(int gold){d_gold = gold;}

        //! Set the defense level of the city
        void setDefenseLevel(int defense_level){d_defense_level = defense_level;}

        /** Set the production of the city
          * 
          * @param production       the index of the internal production slot
          */
        void setProduction(int production);

        //! Set the duration till the production is finished
        void setDuration(int duration){d_duration = duration;}

        /** Set an internal production slot
          * 
          * @param index        the index of the production slot
          * @param armytype     the index of the army within the armyset
          * @param armyset      the armyset of the army, if 0 use the player's
          */
        void setArmytype(int index, int armytype, Armyset* set = 0);

        //! Set if the city is destroyed
        void setBurnt(bool burnt){d_burnt = burnt;}

        //! Sets the production to random starting values
        void setRandomArmytypes();

        //! Produces the currently selected army
        void produceArmy();

        //! Produces the strongest army the city can produce
        void produceLargestArmy();

        //! Add a hero to the stacks within the city
        void addHero(Hero* hero);

        //! Do everything neccessary for a new turn
        void nextTurn();

        //! Returns whether the city can produce a given army type
        bool hasProduction(int type, Armyset* set = 0);

        //! Save the city status. See XML_Helper for info.
        bool save(XML_Helper* helper);
        

        //! Get the unique id of the city
        unsigned int getId() {return d_id;}

        //! Get the defense level of the city
        int getDefenseLevel() const {return d_defense_level;}

        //! Get the player owning the city
        Player* getPlayer() const {return d_player;}

        //! Get the duration till the current production is finished
        int getDuration() const {return d_duration;}

        //! Get the index of the current internal production slot
        int getProduction() const {return d_production;}

        //! Get the income of the city per turn
        int getGold() const {return d_gold;}

        //! Get the index of the army in the given slot
        int getArmytype(int slot) const {return d_armytype[slot].type;}

        //! Get the internal production slot with the given index
        CityProduction getProductiontype(int slot) {return d_armytype[slot];}

        //! Returns whether city was razed (i.e. destroyed)
        bool isBurnt() const {return d_burnt;}

        //! Returns whether the player is a "friend" (==owner) of the city
        bool isFriend(Player* player) const;

    private:
        //! Returns a non-full stack in the city or creates a new one
        Stack* getFreeStack();


        // DATA
        Player* d_player;           // Owner
        
        int d_duration;             // needed turns to finish current production
        CityProduction d_armytype[4];   // possible productiontypes
        int d_production;           // number of produced armytype
        int d_gold;                 // gold produced by city each turn
        bool d_burnt;               // is city burnt down?
        int d_defense_level;        // defense of the city
        unsigned int d_id;
};

#endif // CITY_H
