//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef FIGHT_H
#define FIGHT_H

#include <vector>
#include <deque>
#include <paragui.h>
#include "stack.h"
#include "maptile.h"

class Army;

/** The code for fights is split. While the FightDialog class contains the
  * graphics code and starts new battle rounds, this class does the
  * calculations and contains the whole logic.
  */

class Fight
{ 
    public:
        //! The three possibilities how a fight can end
        enum RESULT {NONE = 0, ATTACKER_WON, DEFENDER_WON};
       
        /** Constructor for a fight between two stacks
          * 
          * This constructor is used if there is just one attacker and one
          * defender stack. The terrain bonus can be explicitely switched off
          * e.g. if a hero searches a ruin, so that the defender has no
          * advantage of being in the mountains.
          *
          * @param maptile          the tile where the fight takes place
          * @param attacker         the attacking stack
          * @param defender         the defending stack
          * @param def_bonus        if true, add terrain bonus to defender
          */
        Fight(Maptile* maptile, Stack* attacker, Stack* defender, bool def_bonus = true);

        /** Constructor for fight between more than two stacks
          * 
          * This is basically the same as the previous constructor, but the
          * attacker and defender stacks are handed over via lists; therefore an
          * arbitrary number of stacks can participate in the fight. The maximum
          * number of stacks is currently limited to 1 attacker and up to 4
          * defenders.
          *
          * @param maptile          the tile where the fight takes place
          * @param attacker         list of the attacking stacks
          * @param defender         list of the defending stacks
          * @param def_bonus        if true, add terrain bonus to defender
          */
        Fight(Maptile* maptile, std::vector<Stack*>& attacker, std::vector<Stack*>& defender,
                bool def_bonus = true);
        ~Fight();


        /** Initialises a new round.
          * 
          * @return false if the maximum number of fight rounds has been
          * exceeded.
          */
        bool newRound();

        /** Lets the current army attack
          * 
          * Calling this function results in the currently active army
          * attacking a random enemy and the active pointer going on to the next
          * army. To give the attacker an advantage, the attacker's armies fight
          * first, then come the defenders (if they aren't dead already).
          * 
          * When all armies have done one attack, a fight round is finished.
          */
        RESULT nextFight();    //do one fight

        //! Returns whether a fight round is completed
        bool isRoundFinished() const {return d_round_finished;}

        
        // CONSTANTS
        static const int ATTACKER_MAX = 1;
        static const int DEFENDER_MAX = 4;
        static const int MAX_ROUNDS = 30;
        static const int MOVES_FOR_FIGHT = 3;

    private:
        /** This function does the real fight between two armies. The attacker
          * tries to land a hit on the defender. First, the strength of the
          * attacker and the defense of the defender is modified with several
          * boni, then hasHit is called to see whether the attacker has
          * succeeded with the attack.
          * 
          * @param culprit      the attacking army
          * @param victim       the defending army
          * @param attack       if culprit is an attacker. Used for several boni
          */
        bool fightArmies(Army* culprit, Army* victim, bool attack);

        //! Calculates the defense bonus of a stack
        void calculateDefenseBonus(Stack* stack);

        //! Checks if there is a hero on the attacker's or defender's side
        void checkHero();

        /** This function calculates if an attacker has landed a hit. It uses a
          * random function. However, it guarantees, that the chanc for an
          * attack is never higher than 95% or lower than 5%. So a throng of
          * pikemen can still kill a dragon, though they are weaklings compared
          * to the beast.
          *
          * @param att_strength         the modified strength of the attacker
          * @param def_strength         the modified defense of the defender
          */
        bool hasHit(float att_strength, float def_strength);

        // DATA
        Maptile* d_mapTile; // maptile on which the fight takes place
        std::deque<Army*> d_attacker;
        std::deque<Army*> d_defender;
        PG_Point d_pos;
        int d_turn;
        float d_add_defense;        //defense bonus of defender
        bool d_open_fight;        //is it a fight on open terrain
        bool d_round_finished;
        bool d_attacker_finished;
        bool d_atthero, d_defhero;  //true if attackers/defenders have a hero
        std::deque<Army*>::iterator d_active;
        RESULT d_result;
};

#endif // FIGHT_H

