//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef MAPTILE_H
#define MAPTILE_H

#include "Tile.h"
#include "TileSet.h"

/** A single tile on the map
  * 
  * The Maptile class encapsulates all information one may want to get about a
  * single maptile of the game map. Specifically, it stores the type of and the
  * buildings on the map tile.
  *
  * A remark concerning the type. A maptile has two types. First, the type value
  * is an index in the tileset which says which tile type this maptile has, e.g.
  * "This maptile is of the sort of the first tile in the tileset". The maptile
  * type says which terrain type this maptile has, e.g. grass or swamps.
  */

class Maptile
{
    public:
        /** Default constructor
          * 
          * @param tileSet          the tileset to use
          * @param x                the x position of the tile
          * @param y                the y position of the tile
          * @param type             the index of the terrain type
          */
        Maptile(TileSet* tileSet, int x, int y, int type);
        ~Maptile();

        //! Enum of possible buldings on the tile
        enum Building {NONE=0, CITY=1, RUIN=2, TEMPLE=3};

        //! Set the type of the terrain (type is an index in the tileset)
        void setType(int type){d_type = type;}

        //! Set which images to use for the corners (see MapRenderer.cpp)
        void setCorners(int c1, int c2, int c3, int c4);

        //! Set the number of moves needed for this maptile
        void setMoves(int moves){d_moves = moves;}

        //! Set which building is on this maptile
        void setBuilding(Building building){d_building = building;}

        
        //! Get the type of the maptile (the index in the tileset)
        int getType(){return d_type;}

        //! Get the image for a corner of the maptile
        int getCorner(int c){return d_corner[c];}

        //! Get which building is on the maptile
        int getBuilding(){return d_building;}

        //! Get the number of moves needed to cross this maptile
        int getMoves(){return d_moves;}

        //! Get the defense bonus for this maptile (only terrain bonus)
        int getDefense(){return d_tileSet->getTile(d_type)->getDefense();}
        
        //! Get the smallmap color of this maptile
        SDL_Color getColor();

        //! Get the tile type (the type of the underlying terrain)
        unsigned int getMaptileType();

        //! Return if the maptile has already been smoothed (soft transitions)
        bool isSmoothed(){return d_smoothed;}

        //prints some debug information
        void printDebugInfo();
                
    private:
        TileSet* d_tileSet;
        short d_type;
        int d_corner[4];        // transition
        short d_moves;          // moves needed to walk over maptile
        Building d_building;    // which building is on this maptile
        bool d_smoothed;
};

#endif // MAPTILE_H

// End of file
