//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef PATH_H
#define PATH_H

#include <set>
#include <map>
#include <list>
#include <paragui.h>
#include "xmlhelper.h"

class PathItem;
class Stack;

/** The path class cares for path storing and calculation.
  * 
  * The task of the path class is three-fold.
  * - it stores existing movement paths of stacks (each stack has a path
  *   instance)
  * - it checks existing paths if they are blocked (when a stack doesn't reach
  *   its target in one round, you have to validate the path in the next round)
  *   and may circumvent obstacles
  * - it calculates the shortest path between points on the map
  *
  * For path calculation we use the A* algorithm. If you are interested how it
  * works, google for A* and path finding. The A* is a modified Dijkstra-Rindaal
  * (maybe the spelling is wrong, I don't have a proper book at hand) algorithm.
  * It should be explained in any better book on graph theory.
  *
  * If you just want to use it, I'll add the shortest possible explanation about
  * how to use it. You can set a heuristic value which determines how much the
  * path moves around obstacles. Set it to zero and the code will review _every_
  * possible path (the original D.-R. algorithm). Set it to high values and the
  * path will be very direct. A reasonable setting for normal use is the value
  * of movement points on grass.
  */

class Path : public std::list<PG_Point*>
{
    public:
        //! Default constructor
        Path();

        //! Loads the existing path from a savegame
        Path(XML_Helper* helper);
        ~Path();

        //! Set the heuristic to some value (default = 2)
        void setHeuristic(double value);

        //! Get the current heuristic value
        double getHeuristic();
        
        /** Calculates the path from the stack's location to a destination.
          *
          * The calculated path is stored within the instance (remember: each
          * stack has a path instance). During calculation, boni and other
          * specialities are taken into account.
          * 
          * @param s            the stack whose path we calculate
          * @param _dest        the destination of the calculation
          * @return true if calculation suceeded or false if path was blocked
          */
        bool calculate(Stack* s, PG_Point _dest);

        /** Validates an existing path
          * 
          * @param s            the Stack whose path we validate
          * @param recalculate  if set to true, try to move around obstacles
          * @return true if path is valid, false if path is blocked (and could
          * not be recalculated)
          */
        bool checkPath(Stack* s, bool recalculate = true);

        /** This function erases an item in the path list. We can't use STL
          * functions, because we need to delete the pointer as well.
          * Otherwise, it behaves like a normal erase function from the STL.
          */
        Path::iterator fl_erase(Path::iterator it);

        /** The same problem. Erases all items in the path list with the help of
          * fl_erase.
          */
        void fl_clear();

        //! Save the current path
        bool save(XML_Helper* helper);

    private:
        //the rest is purely for calculating the path
        bool calculatePath(Stack* s, PG_Point _dest, PG_Point start,
            std::list<PG_Point>* pathlist,
            bool write_directly = false);
        void checkSurroundings(PathItem* p);
        bool isBlocked(int x, int y);
        PathItem createItem(int x, int y, PathItem* parent);
        void insertItem(PathItem item);

        double d_heuristic;
        PG_Point d_dest;
        Stack* d_unit;    //need this for is_blocked()
        unsigned int d_stack_bonus, d_army_bonus;
        bool d_has_ship;
        bool d_hasLand; //has_land refers to units who can't cross water
        std::multiset<PathItem, std::less<PathItem> > d_openlist;
        std::map<unsigned int, PathItem, std::less<unsigned int> > d_closelist;
            //the key is built like key = (x << 16) + y
            //helps with searching (after some cycles closelist
            //contains by far most elements), you only need to find
            //keys Disadvantage: doesn't work on <32bit machines :)
};

#endif // PATH_H
