//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "playerlist.h"
#include <pgmessagebox.h>
#include <sstream>
#include "citylist.h"
#include "defs.h"

using namespace std;

#define debug(x) {cerr<<__FILE__<<": "<<__LINE__<<": "<<x<<endl;}
//#define debug(x)

Playerlist* Playerlist::s_instance = 0;
Player* Playerlist::d_activeplayer = 0;

Playerlist* Playerlist::getInstance()
{
    if (s_instance == 0)
        s_instance = new Playerlist();

    return s_instance;
}

Playerlist* Playerlist::getInstance(XML_Helper* helper)
{
    if (s_instance)
        deleteInstance();

    s_instance = new Playerlist(helper);

    return s_instance;
}

void Playerlist::deleteInstance()
{
    if (s_instance)
    {
        delete s_instance;
        s_instance = 0;
    }
}

Playerlist::Playerlist()
    :d_neutral(0)
{
    d_activeplayer = 0;
}

Playerlist::Playerlist(XML_Helper* helper)
{
    //load data. This consists currently of two values: size (for checking
    //the size; we don't use it yet) and active(which player is active)
    //we do it by calling load with playerlist as string
    load("playerlist", helper);

    helper->registerTag("player", SigC::slot((*this), &Playerlist::load));
}

Playerlist::~Playerlist()
{
    iterator it = begin();
    while (!empty())
        it = fl_erase(it);
}

void Playerlist::checkPlayers()
{
    debug("checkPlayers()");
    iterator it = begin ();
    while (it != end ())
    {
        //ignore the neutral player as well as dead ones
        if (((*it) == d_neutral) || ((*it)->isDead()))
        {
            it++;
            continue;
        }
        
        if (!Citylist::getInstance()->countCities((*it)))
        {
            iterator nextit = it;
            nextit++;

            (*it)->kill();
            splayerDead.emit(*it);

            it = nextit;    // do this at the end to catch abuse of invalid it
        } else {
            ++it;
        }
    }
}

void Playerlist::nextPlayer()
{
    debug("nextPlayer()");

    iterator it;

    if (!d_activeplayer)
    {
        it = begin();
    }
    else
    {
        for (it = begin(); it != end(); ++it)
        {
            if ((*it) == d_activeplayer)
            {
                it++;
                break;
            }
        }
    }
    
    // in case we have got a dead player, continue iterating. This breaks
    // if we have ONLY dead players which we assume never happens.
    while ((it == end()) || ((*it)->isDead()))
    {
        if (it == end())
        {
            it = begin();
            continue;
        }
        it++;
    }

    d_activeplayer = (*it);
    debug("got player: " <<d_activeplayer->getName())
}

Player* Playerlist::getPlayer(string name)
{
    debug("getPlayer()");
    if (name == "Neutral") return d_neutral;
    else
    {
        for (iterator it = begin(); it != end(); ++it)
        {
            if ((*it)->getName() == name) return (*it);
        }
    }
    return 0;
}

Player* Playerlist::getPlayer(unsigned int id)
{
    debug("getPlayer()");
    for (iterator it = begin(); it != end(); it++)
    {
        if (id == (*it)->getId())
            return (*it);
    }

    return 0;
}

int Playerlist::getPlayerNr(Player* p)
{
    int index = 0;
    for (iterator it = begin(); it != end(); ++it, ++index)
    {
        if ((*it) == d_neutral)
            index--;
        if ((*it) == p) return index;
    }
    //-1 == neutral player
    return -1;    
}

int Playerlist::getNoOfPlayers()
{
    int number = 0;
    
    for (iterator it = begin(); it != end(); it++)
    {
        if (((*it) != d_neutral) && !(*it)->isDead())
            number++;
    }

    return number;
}

Player* Playerlist::getFirstLiving()
{
    for (iterator it = begin(); ; it++)
        if (!(*it)->isDead())
            return (*it);
}

bool Playerlist::save(XML_Helper* helper)
{
    //to prevent segfaults
    if (!d_activeplayer)
        d_activeplayer = (*begin());

    bool retval = true;

    retval &= helper->openTag("playerlist");
    retval &= helper->saveData("size", size());    //just to make sure
    retval &= helper->saveData("active", d_activeplayer->getId());

    for (iterator it = begin(); it != end(); it++)
        retval &= (*it)->save(helper);

    retval &= helper->closeTag();

    return retval;
}

bool Playerlist::load(string tag, XML_Helper* helper)
{
    static unsigned int active;
    
    if (tag == "playerlist") //only called in the constructor
    {
        helper->getData(active, "active");
        return true;
    }

    if (tag != "player")
        return false;

    //else tag == "player"
    Player* p = Player::loadPlayer(helper);
    if(p == 0)
        return false;
    
    //insert player...
    push_back(p);

    //set neutral and active
    if (p->getType() == Player::NEUTRAL)
        d_neutral = p;
    if (p->getId() == active)
        d_activeplayer = p;
            
    return true;
}

Playerlist::iterator Playerlist::fl_erase(Playerlist::iterator it)
{
    if ((*it) == d_neutral)
        d_neutral = 0;
    
    delete (*it);
    return erase (it);
}
