//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef REAL_PLAYER_H
#define REAL_PLAYER_H

#include <string>
#include <list>
#include <pgmessageobject.h>
#include "defs.h"
#include "player.h"

#include "signals.h" // include culpritsType

/** This class implements the action function from player.h in a "reasonable"
  * manner for local players. It is suitable fo rlocal human players, AI players
  * can derive from this class and overwrite the start_turn and "callback"
  * functions for their own purposes. I will not describe the single functions
  * here apart from newly introduced ones, see the detailed comments in player.h.
  */

class RealPlayer : public Player
{
    public:
        // CREATORS
        RealPlayer(std::string name, std::string armyset, SDL_Color color);
        RealPlayer(std::string name, std::string armyset, SDL_Color color, Player::TYPE type);
        RealPlayer(const Player&);
        RealPlayer(XML_Helper* helper);
        ~RealPlayer();
        
        //ACTIONS
        virtual bool startTurn();
        virtual bool invadeCity(City* c);
        virtual bool recruitHero(Hero* hero, int cost);
        
        bool stackSplit(Stack* s);
        bool stackJoin(Stack* receiver, Stack* joining);

        bool stackMove(Stack* s);
        bool stackMove(Stack* s, PG_Point dest, bool follow);
        bool stackSearchRuin(Stack* s, Ruin* r);
        bool stackVisitTemple(Stack* s, Temple* t);
        
        Fight::RESULT stackFight(Stack* sattacker, Stack* sdefender);
        Fight::RESULT stackFightCity(std::vector<Stack*> attackers,
                        City* target);

        /** Since fighting a city and fighting a stack is mostly the same, the
          * identical code has been moved to the coreStackFight function. It
          * cares for the fight itself, cleans dead armies, calculates the
          * experience points gained and adds them to the stacks.
          *
          * @param context          one of "city", "ruin", "fight", for setup
          * @param attackers        list of attacking stacks
          * @param defenders        list of defending stacks
          * @param client_data      additional data (e.g. the searched ruin)
          * @return one of Fight::ATTACKER_WON, DEFENDER_WON, NONE
          */
        Fight::RESULT coreStackFight(const char *context, 
                                     std::vector<Stack*>& attackers,
                                     std::vector<Stack*>& defenders,
                                     void *client_data = NULL);

        /** Increases the number of experience points of a stack
          * 
          * This functions takes a number of experience points and distributes
          * them equally over all armies in the stack list. Therefore, the less
          * armies fight, the more experience the single armies get. It can
          * optionally emit a signal if an army advances a level
          *
          * @param stacks           a list of all stacks gaining experience
          * @param emit             if true, emit snewLevelArmy, else ..AiArmy
          * @param xp_sum           the number of XP to distribute
          */
        void updateArmyXp(std::vector<Stack*>& stacks,bool emit,double xp_sum);

        /**
          * Returns all heroes in the stack list
          *
          * @param stacks           the list which is searched
          * @param dst              a list of id's of the heroes found
          */
        void getHeroes(std::vector<Stack*>& stacks, culpritsType& dst);

        /** Removes dead armies in a stack list
          * 
          * @param stacks           the list searched for dead armies
          * @param culprits         the list of armies responsible for killing
          * @return the sum of the XP of the killed armies
          */
        double removeDeadArmies(std::vector<Stack*>& stacks,
                              culpritsType& culprits);

        bool cityOccupy(City* c);
        bool cityPillage(City* c);
        bool cityRaze(City* c);
        bool cityUpgradeDefense(City* c);
        bool cityBuyProduction(City* c, int armytype);
        bool cityChangeProduction(City* c, int production);
        
    protected:
        //if reward != 0, give this reward, else randomize
        bool giveReward (int level, Stack* s, fl_Treasure* reward = 0);
        bool stackMoveOneStep(Stack* s);
        double d_attacker_xp;
        double d_defender_xp;
};

#endif // REAL_PLAYER_H

// End of file
