//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU Library General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#ifndef STACK_H
#define STACK_H

#include <list>
#include "xmlhelper.h"
#include "paragui.h"
#include "signals.h"

class Player;
class Path;
class Army;

/** Group of up to eight armies
  * 
  * While armies are the actual troops you command, they always belong to a
  * stack. The stack holds these armies together in one object. The player
  * usually doesn't command the armies but the stack, so all functionality and
  * data which affects player's controls is bundled in the stack class. Among
  * this is the location of the units, the movement etc.
  */

class Stack : public std::list<Army*>, public SigC::Object
{
    public:
        /** Default constructor
          * 
          * @param player       the owning player or 0 if e.g. ruin keeper
          * @param pos          the position where the stack is created
          */
        Stack(Player* player, PG_Point pos);

        //! Copy constructor
        Stack(Stack& s);

        //! Loading constructor. See XML_Helper for details
        Stack(XML_Helper* helper);
        ~Stack();

        
        /** Sets the defending value. Defending means that this stack is ignored
          * when a player cycles through his list of stacks with Stack::setNext().
          */
        void setDefending(bool defending){d_defending = defending;}

        //! Heals the armies of the stack etc.
        void nextTurn();

        //! Reduces movement points of the armies.
        void decrementMoves(int moves);

        //! Sets the stack's position to the next item of the internal path
        bool moveOneStep();

        //! Resets the moves of all armies to their maximum
        void resetMoves();

        //! Blesses all armies of the stack (strength +1)
        void bless();

        
        //! Returns the unique id of the stack
        unsigned int getId() const {return d_id;}

        //! Returns the owning player
        Player* getPlayer() const {return d_player;}

        //! Returns the location of the stack
        PG_Point getPos() {return d_pos;}

        //! Returns the internal path object of the stack
        Path* getPath() const {return d_path;}

        //! Returns the minimum number of MP of all armies
        int getGroupMoves();

        //! Get the next item of the stack's path
        PG_Point nextStep();

        //! Get the strongest army (most strength) for displaying
        Army* getStrongestArmy() const;

        //! Used for splitting stacks. See Player::stackSplit how it works.
        Army* getFirstUngroupedArmy() const;

        //! True if the stack has a hero. They add strength to the other armies.
        bool hasHero() const;

        //! Return the first hero in the stack
        Army* getFirstHero() const;

        //! Returns the ids of all heroes in the stack in the dst reference
        void getHeroes(culpritsType& dst);

        //! Return true if the stack belongs to player or his allies
        bool isFriend(Player* player) const;

        //! Return the defending status of the stack (see setDefending)
        bool getDefending() const {return d_defending;}

        //! Save the stack's data. See XML_Helper for more details.
        bool save(XML_Helper* helper);

        //! Callback for loading the stack
        bool load(std::string tag, XML_Helper* helper);
    
        //! The same as std::list::clear, but alse frees pointers
        void flClear();

        //! The same as std::list::erase, but also frees pointers
        iterator flErase(iterator object);
        
    private:    
        // DATA
        unsigned int d_id;
        Player* d_player;
        Path* d_path;
        bool d_defending;
        PG_Point d_pos;
};

#endif // STACK_H

// End of file
